@@ -44,6 +44,15 @@
  */
 #define DEFAULT_SERVER "localhost irc.efnet.net irc.undernet.org irc.dalnet.org irc.worldirc.org"
 
+/*
+ * A recent change was to make the %S status line expando always display
+ * the server information, even if you were connected to only one server.
+ * However, a few people expressly dislike this change and want the old 
+ * behavior.  If this is #define'd, you get the old behavior.  If this is
+ * #undef'd, you get the new behavior.
+ */
+#define OLD_STATUS_S_EXPANDO_BEHAVIOR
+
 /*
  * Oooops!  This was supposed to be here, but it never did make it.  Which
  * explains a lot.  Anyhow, #define this if you want to have the option of
@@ -325,6 +334,15 @@
  */
 #undef ANONYMOUS_COMPILE
 
+/*
+ * The /LOAD path is now generated at runtime, rather than at compile time.
+ * This is to allow you to change IRCLIB and have its script library be
+ * resepected without having to change IRCPATH as well.  This is a printf
+ * format of what the default load path is to be.  The %s format indicates
+ * the runtime IRCLIB value.  This value is only used at startup time.
+ */
+#define DEFAULT_IRCPATH "~/.epic:~/.irc:%s/script:."
+
 
 /*
  * Below are the IRCII variable defaults.  For boolean variables, use 1 for
@@ -398,8 +416,11 @@
 #define DEFAULT_LOG 0
 #define DEFAULT_LOGFILE "irc.log"
 #define DEFAULT_MAIL 2
+#define DEFAULT_MAX_RECONNECTS 4
+#define DEFAULT_META_STATES 40
 #define DEFAULT_MODE_STRIPPER 0
 #define DEFAULT_ND_SPACE_MAX 160
+#define DEFAULT_NEW_SERVER_LASTLOG_LEVEL "ALL"
 #define DEFAULT_NO_CTCP_FLOOD 1
 #define DEFAULT_NO_FAIL_DISCONNECT 0
 #define DEFAULT_NOTIFY_HANDLER "NOISY"
@@ -410,7 +431,9 @@
 #define DEFAULT_NO_CONTROL_LOG 0
 #define DEFAULT_NUM_OF_WHOWAS 1
 #define DEFAULT_PAD_CHAR ' '
-#define DEFAULT_QUIT_MESSAGE "ircII %s -- This is the default quit message."
+#define DEFAULT_QUIT_MESSAGE "ircII %s -- Accept no limitations"
+#define DEFAULT_RANDOM_LOCAL_PORTS 0
+#define DEFAULT_RANDOM_SOURCE 0
 #define DEFAULT_REVERSE_STATUS_LINE 1
 #define DEFAULT_SCROLLBACK 256
 #define DEFAULT_SCROLLBACK_RATIO 50
@@ -473,6 +496,7 @@
 #define DEFAULT_SUPPRESS_SERVER_MOTD 0
 #define DEFAULT_TAB 1
 #define	DEFAULT_TAB_MAX 0
+#define DEFAULT_TERM_DOES_BRIGHT_BLINK 0
 #define DEFAULT_UNDERLINE_VIDEO 1
 #define DEFAULT_USERINFO "EPIC4 -- Get your laundry brighter, not just whiter!"
 #define DEFAULT_VERBOSE_CTCP 1
