@@ -1,15 +1,40 @@
 /*
  * alias.c -- Handles the whole kit and caboodle for aliases.
  *
- * Written By Michael Sandrof
- * Copyright 1990, 1995 Michael Sandroff
- *
- * Almost totaly rewritten by Jeremy Nelson (01/97)
- * Copyright 1997 EPIC Software Labs
- *
- * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT
+ * Copyright (c) 1990 Michael Sandroff.
+ * Copyright (c) 1991, 1992 Troy Rollo.
+ * Copyright (c) 1992-1996 Matthew Green.
+ * Copyright  1997, 1998 Jeremy Nelson and others ("EPIC Software Labs").
+ * All rights reserved.
+ * 
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notices, the above paragraph (the one permitting redistribution),
+ *    this list of conditions and the following disclaimer in the 
+ *    documentation and/or other materials provided with the distribution.
+ * 3. The names of the author(s) may not be used to endorse or promote 
+ *    products derived from this software without specific prior written
+ *    permission.
+ * 
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
+ * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
+ * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
+ * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
+ * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
+ * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+ * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
+ * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
+ * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
+ * SUCH DAMAGE.
  */
+/* Almost totaly rewritten by Jeremy Nelson (01/97) */
 
+#define __need_cs_alist_hash__
 #include "irc.h"
 #include "alias.h"
 #include "alist.h"
@@ -55,6 +80,9 @@ static  char *after_expando (char *start, int lvalue, int *call)
 	char	*rest;
 	char	*str;
 
+	if (!*start)
+		return start;
+
 	/*
 	 * One or two leading colons are allowed
 	 */
@@ -63,7 +91,6 @@ static  char *after_expando (char *start, int lvalue, int *call)
 		if (*++str == ':')
 			++str;
 
-
 	/*
 	 * This handles 99.99% of the cases
 	 */
@@ -111,13 +138,15 @@ static  char *after_expando (char *start, int lvalue, int *call)
 	 * If the entire thing looks to be invalid, perhaps its a 
 	 * special built-in expando.  Check to see if it is, and if it
 	 * is, then slurp up the first character as valid.
+	 * Also note that $: by itself must be valid, which requires
+	 * some shenanigans to handle correctly.  Ick.
 	 */
-	if (str == start)
+	if (str == start || (str == start + 1 && *start == ':'))
 	{
 		int is_builtin = 0;
 
 		built_in_alias(*start, &is_builtin);
-		if (is_builtin)
+		if (is_builtin && (str == start))
 			str++;
 	}
 
@@ -140,6 +169,17 @@ static unsigned long 	cmd_cache_misses = 0;
 static unsigned long 	cmd_cache_missed_by = 0;
 static unsigned long 	cmd_cache_passes = 0;
 
+/* Ugh. XXX Bag on the side */
+struct ArgListT {
+	char *	vars[32];
+	char *	defaults[32];
+	int	void_flag;
+	int	dot_flag;
+};
+typedef struct ArgListT ArgList;
+ArgList	*parse_arglist (char *arglist);
+void	destroy_arglist (ArgList *);
+
 
 /*
  * This is where we keep track of where the last pending function call.
@@ -171,7 +211,7 @@ static unsigned long 	cmd_cache_passes = 0;
  */
 extern	void    add_var_alias      (char *name, char *stuff);
 extern  void    add_local_alias    (char *name, char *stuff);
-extern  void    add_cmd_alias      (char *name, char *stuff);
+extern  void    add_cmd_alias      (char *name, ArgList *arglist, char *stuff);
 extern  void    add_var_stub_alias (char *name, char *stuff);
 extern  void    add_cmd_stub_alias (char *name, char *stuff);
 static	void	delete_var_alias   (char *name);
@@ -186,12 +226,15 @@ static	void 	destroy_aliases    (int type);
 
 extern	char *  get_variable            (char *name);
 extern	char ** glob_cmd_alias          (char *name, int *howmany);
+extern	char ** glob_assign_alias	(char *name, int *howmany);
+extern	char ** pmatch_cmd_alias        (char *name, int *howmany);
+extern	char ** pmatch_assign_alias	(char *name, int *howmany);
 extern	char *  get_cmd_alias           (char *name, int *howmany, 
-					 char **complete_name);
+					 char **complete_name, void **args);
 extern	char ** get_subarray_elements   (char *root, int *howmany, int type);
 
 
-static	char *	get_variable_with_args (char *str, char *args, int *args_flag);
+static	char *	get_variable_with_args (const char *str, const char *args, int *args_flag);
 
 /************************** HIGH LEVEL INTERFACE ***********************/
 
@@ -222,6 +265,14 @@ BUILT_IN_COMMAND(aliascmd)
 	 */
 	if (!my_strnicmp(name, "/S", 2))
 	{
+extern u_32int_t       bin_ints;
+extern u_32int_t       lin_ints;
+extern u_32int_t       bin_chars;
+extern u_32int_t       lin_chars;
+extern u_32int_t       alist_searches;
+extern u_32int_t       char_searches;
+
+
 	    say("Total aliases handled: %ld", 
 			alias_total_allocated);
 	    say("Total bytes allocated to aliases: %ld", 
@@ -239,7 +290,9 @@ BUILT_IN_COMMAND(aliascmd)
 			cmd_cache_passes,
 			( cmd_cache_misses ? (double)
 			  (cmd_cache_missed_by / cmd_cache_misses) : 0.0));
-
+	    say("Ints(bin/lin)/Chars(bin/lin)/Lookups: [(%d/%d)/(%d/%d)] (%d/%d)",
+			 bin_ints, lin_ints, bin_chars, lin_chars,
+			 alist_searches, char_searches);
 	    show_alias_caches();
 	    return;
 	}
@@ -284,6 +337,35 @@ BUILT_IN_COMMAND(aliascmd)
 	 */
 	else
 	{
+		ArgList *arglist = NULL;
+
+		/*
+		 * Aliases may contain a parameter list, which is parsed
+		 * at registration time.
+		 */
+		if (*args == LEFT_PAREN)
+		{
+			ptr = MatchingBracket(++args, LEFT_PAREN, RIGHT_PAREN);
+			if (!ptr)
+				say("Unmatched lparen in %s %s", 
+					command, real_name);
+			else
+			{
+				*ptr++ = 0;
+				while (*ptr && my_isspace(*ptr))
+					ptr++;
+				if (!*ptr)
+					say("Missing alias body in %s %s", 
+						command, real_name);
+
+				while (*args && my_isspace(*args))
+					args++;
+
+				arglist = parse_arglist(args);
+				args = ptr;
+			}
+		}
+
 		/*
 		 * Aliases' bodies can be surrounded by a set of braces,
 		 * which are stripped off.
@@ -293,7 +375,8 @@ BUILT_IN_COMMAND(aliascmd)
 			ptr = MatchingBracket(++args, LEFT_BRACE, RIGHT_BRACE);
 
 			if (!ptr)
-				say("Unmatched brace in %s", command);
+				say("Unmatched brace in %s %s", 
+						command, real_name);
 			else 
 			{
 				*ptr++ = 0;
@@ -301,7 +384,7 @@ BUILT_IN_COMMAND(aliascmd)
 					ptr++;
 
 				if (*ptr)
-					say("Junk [%s] after closing brace in %s", ptr, command);
+					say("Junk [%s] after closing brace in %s %s", ptr, command, real_name);
 
 				while (*args && my_isspace(*args))
 					args++;
@@ -312,7 +395,7 @@ BUILT_IN_COMMAND(aliascmd)
 		/*
 		 * Register the alias
 		 */
-		add_cmd_alias(real_name, args);
+		add_cmd_alias(real_name, arglist, args);
 	}
 
 	new_free(&real_name);
@@ -565,6 +648,154 @@ BUILT_IN_COMMAND(unloadcmd)
 }
 
 
+/*
+ * Argument lists look like this:
+ *
+ * LIST   := LPAREN TERM [COMMA TERM] RPAREN)
+ * LPAREN := '('
+ * RPAREN := ')'
+ * COMMA  := ','
+ * TERM   := LVAL QUAL | "..." | "void"
+ * LVAL   := <An alias name>
+ * QUAL   := NUM "words"
+ *
+ * In English:
+ *   An argument list is a comma seperated list of variable descriptors (TERM)
+ *   enclosed in a parenthesis set.  Each variable descriptor may contain any
+ *   valid alias name followed by an action qualifier, or may be the "..."
+ *   literal string, or may be the "void" literal string.  If a variable
+ *   descriptor is an alias name, then that positional argument will be removed
+ *   from the argument list and assigned to that variable.  If the qualifier
+ *   specifies how many words are to be taken, then that many are taken.
+ *   If the variable descriptor is the literal string "...", then all argument
+ *   list parsing stops there and all remaining alias arguments are placed 
+ *   into the $* special variable.  If the variable descriptor is the literal
+ *   string "void", then the balance of the remaining alias arguments are 
+ *   discarded and $* will expand to the false value.  If neither "..." nor
+ *   "void" are provided in the argument list, then that last variable in the
+ *   list will recieve all of the remaining arguments left at its position.
+ * 
+ * Examples:
+ *
+ *   # This example puts $0 in $channel, $1 in $mag, and $2- in $nicklist.
+ *   /alias opalot (channel, mag, nicklist) {
+ *	fe ($nicklist) xx yy zz {
+ *	    mode $channel ${mag}ooo $xx $yy $zz
+ *	}
+ *   }
+ *
+ *   # This example puts $0 in $channel, $1 in $mag, and the new $* is old $2-
+ *   /alias opalot (channel, mag, ...) {
+ *	fe ($*) xx yy zz {
+ *	    mode $channel ${mag}ooo $xx $yy $zz
+ *	}
+ *   }
+ *
+ *   # This example throws away any arguments passed to this alias
+ *   /alias booya (void) { echo Booya! }
+ */
+ArgList	*parse_arglist (char *arglist)
+{
+	char *	this_term;
+	char *	next_term;
+	char *	varname;
+	char *	modifier, *value;
+	int	arg_count = 0;
+	ArgList *args = new_malloc(sizeof(ArgList));
+
+	args->void_flag = args->dot_flag = 0;
+	for (this_term = arglist; *this_term; this_term = next_term,arg_count++)
+	{
+		while (isspace(*this_term))
+			this_term++;
+		next_in_comma_list(this_term, &next_term);
+		if (!(varname = next_arg(this_term, &this_term)))
+			continue;
+		if (!my_stricmp(varname, "void")) {
+			args->void_flag = 1;
+			break;
+		} else if (!my_stricmp(varname, "...")) {
+			args->dot_flag = 1;
+			break;
+		} else {
+			args->vars[arg_count] = m_strdup(varname);
+			args->defaults[arg_count] = NULL;
+
+			if (!(modifier = next_arg(this_term, &this_term)))
+				continue;
+			if (!my_stricmp(modifier, "default"))
+			{
+			    if (!(value = new_next_arg(this_term, &this_term)))
+				continue;
+			    args->defaults[arg_count] = m_strdup(value);
+			}
+		}
+	}
+	args->vars[arg_count] = NULL;
+	return args;
+}
+
+void	destroy_arglist (ArgList *arglist)
+{
+	int	i = 0;
+
+	if (!arglist)
+		return;
+
+	for (i = 0; ; i++)
+	{
+		if (!arglist->vars[i])
+			break;
+		new_free(&arglist->vars[i]);
+		new_free(&arglist->defaults[i]);
+	}
+	new_free((char **)&arglist);
+}
+
+void	prepare_alias_call (void *al, char **stuff)
+{
+	ArgList *args = (ArgList *)al;
+	int	i;
+
+	if (!args)
+		return;
+
+	for (i = 0; args->vars[i]; i++)
+	{
+		char	*next_val;
+		char	*expanded = NULL;
+		int	af;
+
+		/* Last argument on the list and no ... argument? */
+		if (!args->vars[i + 1] && !args->dot_flag && !args->void_flag)
+		{
+			next_val = *stuff;
+			*stuff = empty_string;
+		}
+
+		/* Yank the next word from the arglist */
+		else
+			next_val = next_arg(*stuff, stuff);
+
+		if (!next_val || !*next_val)
+		{
+			if ((next_val = args->defaults[i]))
+				next_val = expanded = expand_alias(next_val, *stuff, &af, NULL);
+			else
+				next_val = empty_string;
+		}
+
+		/* Add the local variable */
+		add_local_alias(args->vars[i], next_val);
+		if (expanded)
+			new_free(&expanded);
+	}
+
+	/* Throw away rest of args if wanted */
+	if (args->void_flag)
+		*stuff = empty_string;
+}
+
 /**************************** INTERMEDIATE INTERFACE *********************/
 /* We define Alias here to keep it encapsulated */
 /*
@@ -574,12 +805,14 @@ BUILT_IN_COMMAND(unloadcmd)
 typedef	struct	AliasItemStru
 {
 	char	*name;			/* name of alias */
+	u_32int_t hash;			/* Hash of the name */
 	char	*stuff;			/* what the alias is */
 	char	*stub;			/* the file its stubbed to */
 	char	*filename;		/* file it was loaded from */
 	int	line;			/* line it was loaded from */
 	int	global;			/* set if loaded from `global' */
 	int	cache_revoked;		/* Cache revocation index */
+	ArgList *arglist;		/* List of arguments to alias */
 }	Alias;
 
 /*
@@ -615,7 +848,7 @@ static	Alias *	find_local_alias   (char *name, AliasSet **list);
  */
 typedef struct RuntimeStackStru
 {
-	char	*name;		/* Name of the stack */
+	const char *name;	/* Name of the stack */
 	char 	*current;	/* Current cmd being executed */
 	AliasSet alias;		/* Local variables */
 	int	locked;		/* Are we locked in a wait? */
@@ -662,6 +895,7 @@ Alias *make_new_Alias (char *name)
 	tmp->line = current_line();
 	tmp->cache_revoked = 0;
 	tmp->filename = m_strdup(current_package());
+	tmp->arglist = NULL;
 	alias_total_bytes_allocated += sizeof(Alias) + strlen(name) +
 				strlen(tmp->filename);
 	return tmp;
@@ -806,7 +1040,7 @@ void	add_local_alias	(char *name, char *stuff)
 	return;
 }
 
-void	add_cmd_alias	(char *name, char *stuff)
+void	add_cmd_alias	(char *name, ArgList *arglist, char *stuff)
 {
 	Alias *tmp = NULL;
 	int cnt, af, loc;
@@ -825,6 +1059,7 @@ void	add_cmd_alias	(char *name, char *stuff)
 	malloc_strcpy(&(tmp->stuff), stuff);
 	new_free(&tmp->stub);
 	tmp->global = loading_global;
+	tmp->arglist = arglist;
 
 	alias_total_allocated++;
 	alias_total_bytes_allocated += strlen(tmp->stuff);
@@ -867,7 +1102,7 @@ void	add_var_stub_alias  (char *name, char *stuff)
 		tmp->global = loading_global;
 
 		alias_total_allocated++;
-		alias_total_bytes_allocated += strlen(tmp->stuff);
+		alias_total_bytes_allocated += strlen(tmp->stub);
 		say("Assign %s stubbed to file %s", name, stuff);
 	}
 
@@ -935,6 +1170,8 @@ static Alias *	find_var_alias (char *name)
 	int 	loc;
 	int 	cnt = 0;
 	int 	i;
+	u_32int_t	mask;
+	u_32int_t	hash = cs_alist_hash(name, &mask);
 
 	if (!strncmp(name, "::", 2))
 		name += 2;		/* Accept ::global */
@@ -944,9 +1181,9 @@ static Alias *	find_var_alias (char *name)
 
 	for (cache = 0; cache < var_alias.cache_size; cache++)
 	{
-		if (var_alias.cache[cache] && 
-		    var_alias.cache[cache]->name &&
-		    !strcmp(name, var_alias.cache[cache]->name))
+		if (var_alias.cache[cache] && var_alias.cache[cache]->name &&
+			(var_alias.cache[cache]->hash == hash) &&
+			!strcmp(name, var_alias.cache[cache]->name))
 		{
 			item = var_alias.cache[cache];
 			cnt = -1;
@@ -983,6 +1220,8 @@ static Alias *	find_var_alias (char *name)
 		if (item->stub)
 		{
 			unstub_alias(item);
+			if (!(item = find_var_alias(name)))
+				return NULL;
 			if (!item->stuff)
 			{
 				int owd = window_display;
@@ -1001,29 +1240,31 @@ static Alias *	find_var_alias (char *name)
 
 static Alias *	find_cmd_alias (char *name, int *cnt)
 {
-	Alias *item = NULL;
-	int loc;
-	int i;
-	int cache;
+	Alias *		item = NULL;
+	int 		loc;
+	int 		i;
+	int 		cache;
+	u_32int_t	mask;
+	u_32int_t	hash = cs_alist_hash(name, &mask);
 
 	if (cmd_alias.cache_size == 0)
 		resize_cache(&cmd_alias, ALIAS_CACHE_SIZE);
 
 	for (cache = 0; cache < cmd_alias.cache_size; cache++)
 	{
-		if (cmd_alias.cache[cache])
+		if (cmd_alias.cache[cache] && cmd_alias.cache[cache]->name &&
+			(cmd_alias.cache[cache]->hash == hash) &&
+			!strcmp(name, cmd_alias.cache[cache]->name))
 		{
-			/* XXXX this is bad cause this is free()d! */
-			if (!cmd_alias.cache[cache]->name)
-				cmd_alias.cache[cache] = NULL;
-			else if (!strcmp(name, cmd_alias.cache[cache]->name))
-			{
-				item = cmd_alias.cache[cache];
-				*cnt = -1;
-				cmd_cache_hits++;
-				break;
-			}
+			item = cmd_alias.cache[cache];
+			*cnt = -1;
+			cmd_cache_hits++;
+			break;
 		}
+
+		/* XXXX this is bad cause this is free()d! */
+		if (cmd_alias.cache[cache] && !cmd_alias.cache[cache]->name)
+			cmd_alias.cache[cache] = NULL;
 	}
 
 	if (!item)
@@ -1054,6 +1295,10 @@ static Alias *	find_cmd_alias (char *name, int *cnt)
 		if (item->stub)
 		{
 			unstub_alias(item);
+			/* Refetch it, it might have gone away */
+			if (!(item = find_cmd_alias(name, cnt)))
+				return NULL;
+
 			if (!item->stuff)
 			{
 				int owd = window_display;
@@ -1079,8 +1324,7 @@ static void 	unstub_alias (Alias *item)
 	static int already_looking = 0;
 	char *copy;
 
-	copy = alloca(strlen(item->stub) + 1);
-	strcpy(copy, item->stub);
+	copy = LOCAL_COPY(item->stub);
 	new_free((void **)&item->stub);
 
 	/* 
@@ -1254,6 +1498,7 @@ void	delete_cmd_alias (char *name)
 		new_free(&(item->stuff));
 		new_free(&(item->stub));
 		new_free(&(item->filename));
+		destroy_arglist(item->arglist);
 		new_free((char **)&item);
 		say("Alias	%s removed", name);
 	}
@@ -1456,7 +1701,7 @@ char 	*get_variable 	(char *str)
 
 
 static 
-char	*get_variable_with_args (char *str, char *args, int *args_flag)
+char	*get_variable_with_args (const char *str, const char *args, int *args_flag)
 {
 	Alias	*alias = NULL;
 	char	*ret = NULL;
@@ -1470,8 +1715,9 @@ char	*get_variable_with_args (char *str, char *args, int *args_flag)
 	/*
 	 * Support $:var to mean local variable ONLY (no globals)
 	 * Support $::var to mean global variable ONLY (no locals)
+	 * Support $: with nothing after it as a built in expando.  Ick.
 	 */
-	if (*name == ':')
+	if (*name == ':' && name[1])
 	{
 		name++, local = 1;
 		if (*name == ':')
@@ -1503,7 +1749,7 @@ char	*get_variable_with_args (char *str, char *args, int *args_flag)
 	return (copy ? m_strdup(ret) : ret);
 }
 
-char *	get_cmd_alias (char *name, int *howmany, char **complete_name)
+char *	get_cmd_alias (char *name, int *howmany, char **complete_name, void **args)
 {
 	Alias *item;
 
@@ -1511,6 +1757,8 @@ char *	get_cmd_alias (char *name, int *howmany, char **complete_name)
 	{
 		if (complete_name)
 			malloc_strcpy(complete_name, item->name);
+		if (args)
+			*args = (void *)item->arglist;
 		return item->stuff;
 	}
 	return NULL;
@@ -1558,6 +1806,121 @@ char **	glob_cmd_alias (char *name, int *howmany)
 	return matches;
 }
 
+/*
+ * This function is strictly O(N).  This should probably be addressed.
+ */
+char **	glob_assign_alias (char *name, int *howmany)
+{
+	int    	cnt;
+	int     cmp;
+	int     len;
+	char    **matches = NULL;
+	int     matches_size = 5;
+
+	len = strlen(name);
+	*howmany = 0;
+	matches = RESIZE(matches, char *, matches_size);
+
+	for (cnt = 0; cnt < var_alias.max; cnt++)
+	{
+		if (!(cmp = strncmp(name, var_alias.list[cnt]->name, len)))
+		{
+			if (strchr(var_alias.list[cnt]->name + len, '.'))
+				continue;
+
+			matches[*howmany] = m_strdup(var_alias.list[cnt]->name);
+			*howmany += 1;
+			if (*howmany == matches_size)
+			{
+				matches_size += 5;
+				RESIZE(matches, char *, matches_size);
+			}
+		}
+		else if (cmp < 0)
+			break;
+	}
+
+	if (*howmany)
+		matches[*howmany] = NULL;
+	else
+		new_free((char **)&matches);
+
+	return matches;
+}
+
+/*
+ * This function is strictly O(N).  This should probably be addressed.
+ */
+char **	pmatch_cmd_alias (char *name, int *howmany)
+{
+	int	cnt;
+	int 	len;
+	char 	**matches = NULL;
+	int 	matches_size = 5;
+
+	len = strlen(name);
+	*howmany = 0;
+	matches = RESIZE(matches, char *, matches_size);
+
+	for (cnt = 0; cnt < cmd_alias.max; cnt++)
+	{
+		if (wild_match(name, cmd_alias.list[cnt]->name))
+		{
+			matches[*howmany] = m_strdup(cmd_alias.list[cnt]->name);
+			*howmany += 1;
+			if (*howmany == matches_size)
+			{
+				matches_size += 5;
+				RESIZE(matches, char *, matches_size);
+			}
+		}
+	}
+
+	if (*howmany)
+		matches[*howmany] = NULL;
+	else
+		new_free((char **)&matches);
+
+	return matches;
+}
+
+/*
+ * This function is strictly O(N).  This should probably be addressed.
+ */
+char **	pmatch_assign_alias (char *name, int *howmany)
+{
+	int    	cnt;
+	int     len;
+	char    **matches = NULL;
+	int     matches_size = 5;
+
+	len = strlen(name);
+	*howmany = 0;
+	matches = RESIZE(matches, char *, matches_size);
+
+	for (cnt = 0; cnt < var_alias.max; cnt++)
+	{
+		if (wild_match(name, var_alias.list[cnt]->name))
+		{
+			matches[*howmany] = m_strdup(var_alias.list[cnt]->name);
+			*howmany += 1;
+			if (*howmany == matches_size)
+			{
+				matches_size += 5;
+				RESIZE(matches, char *, matches_size);
+			}
+		}
+	}
+
+	if (*howmany)
+		matches[*howmany] = NULL;
+	else
+		new_free((char **)&matches);
+
+	return matches;
+}
+
+
 /*
  * This function is strictly O(N).  This should probably be addressed.
  */
@@ -1614,17 +1977,21 @@ char **	get_subarray_elements (char *root, int *howmany, int type)
 }
 
 
-char *	parse_line_with_return (char *name, char *what, char *args, int d1, int d2)
+/* XXX What is this doing here? */
+static char *	parse_line_alias_special (char *name, char *what, char *args, int d1, int d2, void *arglist, int function)
 {
 	int	old_window_display = window_display;
 	int	old_last_function_call_level = last_function_call_level;
-	char	*result;
-
-	make_local_stack(name);
-	last_function_call_level = wind_index;
+	char	*result = NULL;
 
 	window_display = 0;
-	add_local_alias("FUNCTION_RETURN", empty_string);
+	make_local_stack(name);
+	prepare_alias_call(arglist, &args);
+	if (function)
+	{
+		last_function_call_level = wind_index;
+		add_local_alias("FUNCTION_RETURN", empty_string);
+	}
 	window_display = old_window_display;
 
 	will_catch_return_exceptions++;
@@ -1632,13 +1999,20 @@ char *	parse_line_with_return (char *name, char *what, char *args, int d1, int d
 	will_catch_return_exceptions--;
 	return_exception = 0;
 
-	result = get_variable("FUNCTION_RETURN");
-	last_function_call_level = old_last_function_call_level;
+	if (function)
+	{
+		result = get_variable("FUNCTION_RETURN");
+		last_function_call_level = old_last_function_call_level;
+	}
 	destroy_local_stack();
-
 	return result;
 }
 
+char *	parse_line_with_return (char *name, char *what, char *args, int d1, int d2)
+{
+	return parse_line_alias_special(name, what, args, d1, d2, NULL, 0);
+}
+
 /************************************************************************/
 /*
  * call_user_function: Executes a user alias (by way of parse_command.
@@ -1652,10 +2026,12 @@ char 	*call_user_function	(char *alias_name, char *args)
 	char 	*result = NULL;
 	char 	*sub_buffer;
 	int 	cnt;
+	void	*arglist = NULL;
 
-	sub_buffer = get_cmd_alias(alias_name, &cnt, NULL);
+	sub_buffer = get_cmd_alias(alias_name, &cnt, NULL, &arglist);
 	if (cnt < 0)
-		result = parse_line_with_return(alias_name, sub_buffer, args, 0, 1);
+		result = parse_line_alias_special(alias_name, sub_buffer, 
+						args, 0, 1, arglist, 1);
 	else if (x_debug & DEBUG_UNKNOWN)
 		yell("Function call to non-existant alias [%s]", alias_name);
 
@@ -1665,6 +2041,14 @@ char 	*call_user_function	(char *alias_name, char *args)
 	return result;
 }
 
+/* XXX Ugh. */
+void	call_user_alias	(char *alias_name, char *alias_stuff, char *args, void *arglist)
+{
+	parse_line_alias_special(alias_name, alias_stuff, args, 
+					0, 1, arglist, 0);
+}
+
+
 /*
  * save_aliases: This will write all of the aliases to the FILE pointer fp in
  * such a way that they can be read back in using LOAD or the -l switch 
@@ -1723,7 +2107,7 @@ void 	destroy_aliases (int type)
 
 /******************* RUNTIME STACK SUPPORT **********************************/
 
-void 	make_local_stack 	(char *name)
+void 	make_local_stack 	(const char *name)
 {
 	wind_index++;
 
@@ -1864,6 +2248,7 @@ void 	destroy_call_stack 	(void)
 
 /******************* expression and text parsers ***************************/
 /* XXXX - bogus for now */
+#include "expr2.c"
 #include "expr.c"
 
 
@@ -1883,7 +2268,7 @@ char 	*aliasctl 	(char *input)
 {
 	int list = -1;
 	char *listc;
-	enum { GET, SET, MATCH } op;
+	enum { GET, SET, MATCH, PMATCH } op;
 
 	GET_STR_ARG(listc, input);
 	if (!my_strnicmp(listc, "AS", 2))
@@ -1902,6 +2287,8 @@ char 	*aliasctl 	(char *input)
 		op = SET;
 	else if (!my_strnicmp(listc, "M", 1))
 		op = MATCH;
+	else if (!my_strnicmp(listc, "P", 1))
+		op = PMATCH;
 	else
 		RETURN_EMPTY;
 
@@ -1935,7 +2322,7 @@ char 	*aliasctl 	(char *input)
 			else if (list == VAR_ALIAS)
 				add_var_alias(listc, input);
 			else
-				add_cmd_alias(listc, input);
+				add_cmd_alias(listc, NULL, input);
 
 			RETURN_INT(1)
 		}
@@ -1953,8 +2340,29 @@ char 	*aliasctl 	(char *input)
 			if (list == COMMAND_ALIAS)
 				mlist = glob_cmd_alias(listc, &num);
 			else
-/* XXXXXXXXXXX */
-				error("aliasctl: MATCH isnt supported on that yet.  Ask me to support it.");
+				mlist = glob_assign_alias(listc, &num);
+
+			for (ctr = 0; ctr < num; ctr++)
+			{
+				m_s3cat(&mylist, space, mlist[ctr]);
+				new_free((char **)&mlist[ctr]);
+			}
+			new_free((char **)&mlist);
+			if (mylist)
+				return mylist;
+			RETURN_EMPTY;
+		}
+		case (PMATCH) : 
+		{
+			char **	mlist = NULL;
+			char *	mylist = NULL;
+			int	num = 0,
+				ctr;
+
+			if (list == COMMAND_ALIAS)
+				mlist = pmatch_cmd_alias(listc, &num);
+			else
+				mlist = pmatch_assign_alias(listc, &num);
 
 			for (ctr = 0; ctr < num; ctr++)
 			{
