@@ -161,13 +161,14 @@ struct	deadlist
 	struct deadlist *next;
 }	*deadlist = NULL;
 
+int	dcc_dead (void) { return deadlist ? 1 : 0; }
 
 /*
  * remove_from_dcc_list: What do you think it does?
  */
 static void	dcc_remove_from_list (DCC_list *erased)
 {
-	DCC_list *prev;
+	DCC_list *prev = NULL;
 
 	if (erased != ClientList)
 	{
@@ -197,6 +198,10 @@ static void 	dcc_add_deadclient (DCC_list *client)
 {
 	struct deadlist *new_dl;
 
+	for (new_dl = deadlist; new_dl; new_dl = new_dl->next)
+		if (new_dl->it == client)
+			return;
+
 	new_dl = new_malloc(sizeof(struct deadlist));
 	new_dl->next = deadlist;
 	new_dl->it = client;
@@ -235,6 +240,7 @@ static 	void		dcc_erase (DCC_list *erased)
 		unsigned my_type = erased->flags & DCC_TYPES;
 		char	*dummy_ptr = NULL;
 		char 	*nopath;
+		char	*dummy_nick = NULL;
 
 		if (erased->description && 
 				(nopath = strrchr(erased->description, '/')))
@@ -243,10 +249,12 @@ static 	void		dcc_erase (DCC_list *erased)
 			nopath = erased->description;
 
 		if (*(erased->user) == '=')
-			erased->user++;
+			dummy_nick = erased->user + 1;
+		else
+			dummy_nick = erased->user;
 
 		dummy_ptr = m_sprintf("%s %s %s", 
-			erased->user, 
+			dummy_nick,
 			(my_type == DCC_FILEOFFER ? "GET" :
 			 (my_type == DCC_FILEREAD  ? "SEND" : 
 			   dcc_types[my_type])),
@@ -272,15 +280,13 @@ static 	void		dcc_erase (DCC_list *erased)
 	erased->write = new_close(erased->write);
 	erased->read = new_close(erased->read);
 	erased->file = new_close(erased->file);
-	new_free(&erased->description);
+	new_free(&erased->description);	/* Magic check failure here */
 	new_free(&erased->filename);
 	new_free(&erased->user);
 	new_free(&erased->othername);
 	new_free(&erased->encrypt);
 	new_free(&erased->cksum);
 	new_free((char **)&erased);
-
-	
 }
 
 /*
@@ -615,16 +621,27 @@ static	int		dcc_open (DCC_list *dcc)
 	 */
 	else
 	{
-		unsigned short 	portnum = dcc->local_port;
+		unsigned short 	portnum1 = 0;
+		unsigned short	portnum2 = 0;
+
+		if (get_int_var(RANDOM_LOCAL_PORTS_VAR))
+			portnum1 = random_number(65535 - 1024) + 1024;
 
 		/*
 		 * Mark that we're waiting for the remote peer to answer,
-		 * and then open up a listen()ing socket for them.
+		 * and then open up a listen()ing socket for them.  If our
+		 * first choice of port fails, try another one.  If both
+		 * fail, then we give up.  If the user insists on doing
+		 * random ports, then we will fallback to asking the system
+		 * for a port if our random port isnt available.
 		 */
 		dcc->flags |= DCC_MY_OFFER;
-		if ((dcc->read = connect_by_number(NULL, &portnum, 
+		if ((dcc->read = connect_by_number(NULL, &portnum1, 
 					SERVICE_SERVER, PROTOCOL_TCP)) < 0)
 		{
+		    if ((dcc->read = connect_by_number(NULL, &portnum2,
+					SERVICE_SERVER, PROTOCOL_TCP)) < 0)
+		    {
 			message_from(NULL, LOG_DCC);
 			say("Unable to create connection: %s", 
 				errno ? strerror(errno) : "Unknown Host");
@@ -632,7 +649,13 @@ static	int		dcc_open (DCC_list *dcc)
 			dcc->flags |= DCC_DELETE;
 			from_server = old_server;
 			return 0;
+		    }
+		    else
+			dcc->local_port = portnum2;
 		}
+		else
+		    dcc->local_port = portnum1;
+
 
 #ifdef MIRC_BROKEN_DCC_RESUME
 		/*
@@ -641,10 +664,9 @@ static	int		dcc_open (DCC_list *dcc)
 		 * back that port number as its ID of what file it wants
 		 * to resume (rather than the filename. ick.)
 		 */
-		malloc_strcpy(&dcc->othername, ltoa(portnum));
+		malloc_strcpy(&dcc->othername, ltoa(dcc->local_port));
 #endif
 		new_open(dcc->read);
-		dcc->local_port = portnum;
 
 		/*
 		 * If this is to be a 2-peer connection, then we need to
@@ -671,7 +693,7 @@ static void	dcc_send_booster_ctcp (DCC_list *dcc)
 	char		*type = dcc_types[dcc->flags & DCC_TYPES];
 	IA		myip;
 
-	myip.s_addr = server_list[from_server].local_addr.s_addr;
+	myip.s_addr = get_server_local_addr(from_server).s_addr;
 
 	/*
 	 * If this is to be a 2-peer connection, then we need to
@@ -754,9 +776,9 @@ static void	dcc_send_booster_ctcp (DCC_list *dcc)
 		 */
 		message_from(NULL, LOG_DCC);
 		if (do_hook(DCC_OFFER_LIST, "%s %s", 
-			dcc->user, type))
+				dcc->user, type))
 		    say("Sent DCC %s request to %s", 
-			type, dcc->user);
+				type, dcc->user);
 		message_from(NULL, LOG_CURRENT);
 	}
 }
@@ -783,7 +805,7 @@ static void	dcc_message_transmit (
 	char 		*text, 		/* What to send them */
 const	char 		*text_display, 	/* What to tell the user we sent */
 	int 		noisy, 		/* Do we tell the user? */
-	char 		*cmd)		/* 
+	const char 	*cmd)		/* 
 					 * For /CTCP's, is this a PRIVMSG
 					 * or a NOTICE?
 					 */
@@ -868,7 +890,7 @@ const	char 		*text_display, 	/* What to tell the user we sent */
  * This is used to send a message to a remote DCC peer.  This is called
  * by send_text.
  */
-void	dcc_chat_transmit (char *user, char *text, const char *orig, char *type, int noisy)
+void	dcc_chat_transmit (char *user, char *text, const char *orig, const char *type, int noisy)
 {
 	int	fd;
 
@@ -929,7 +951,20 @@ void	process_dcc(char *args)
 			message_from(NULL, LOG_DCC);
 			dcc_commands[i].function(args);
 			message_from(NULL, LOG_CURRENT);
-			dcc_check(NULL);	/* Clean up */
+#if 0
+			/*
+			 * If the user called /dcc close, and we were
+			 * called by do_hook() via an /on, then it may be
+			 * very possible that somewhere up the stack is
+			 * someone who is expecting their DCC client pointer
+			 * to be valid when do_hook() gets back.  Rather than
+			 * "fixing" all those callers, we just try not to be
+			 * so aggressive about garbage collecting.  We instead
+			 * defer to the next call to dcc_check() which will
+			 * happen no later than the top of the next minute.
+			 */
+			dcc_check(NULL);	/* Clean up - XXX Wrong */
+#endif
 			return;
 		}
 	}
@@ -944,6 +979,7 @@ static void	dcc_chat (char *args)
 {
 	char		*user;
 	DCC_list	*dcc;
+	int		portnum = 0;
 
 	if ((user = next_arg(args, &args)) == NULL)
 	{
@@ -951,6 +987,18 @@ static void	dcc_chat (char *args)
 		return;
 	}
 
+	/*
+	 * Check to see if there is a flag
+	 */
+	if (*args == '-')
+	{
+		if (args[1] == 'p')
+		{
+			if (args && *args)
+			    portnum = my_atol(next_arg(args, &args));
+		}
+	}
+
 	dcc = dcc_searchlist("chat", user, DCC_CHAT, 1, NULL, -1);
 	if ((dcc->flags & DCC_ACTIVE) || (dcc->flags & DCC_MY_OFFER))
 	{
@@ -959,6 +1007,7 @@ static void	dcc_chat (char *args)
 	}
 
 	dcc->flags |= DCC_TWOCLIENTS;
+	dcc->local_port = portnum;
 	dcc_open(dcc);
 }
 
@@ -1445,7 +1494,7 @@ static	void	dcc_filesend (char *args)
 		}
 
 		Client->flags |= DCC_TWOCLIENTS;
-		Client->remport = portnum;
+		Client->local_port = portnum;
 		dcc_open(Client);
 	    } /* The WHILE */
 	} /* The IF */
@@ -1469,7 +1518,7 @@ char	*dcc_raw_listen (unsigned short port)
 	{
 		say("May not bind to a privileged port");
 		message_from(NULL, LOG_CURRENT);
-		return NULL;
+		return m_strdup(empty_string);
 	}
 	PortName = ltoa(port);
 	Client = dcc_searchlist("raw_listen", PortName, 
@@ -1479,7 +1528,7 @@ char	*dcc_raw_listen (unsigned short port)
 	{
 		say("A previous DCC RAW_LISTEN on %s exists", PortName);
 		message_from(NULL, LOG_CURRENT);
-		return NULL;
+		return m_strdup(empty_string);
 	}
 
 	if ((Client->read = connect_by_number(NULL, &port, SERVICE_SERVER, PROTOCOL_TCP)) < 0)
@@ -1487,7 +1536,7 @@ char	*dcc_raw_listen (unsigned short port)
 		say("Couldnt establish listening socket: [%d] %s", Client->read, strerror(errno));
 		Client->flags |= DCC_DELETE; 
 		message_from(NULL, LOG_CURRENT);
-		return NULL;
+		return m_strdup(empty_string);
 	}
 
 	new_open(Client->read);
@@ -1620,7 +1669,11 @@ void	register_dcc_offer (char *user, char *type, char *description, char *addres
 	Client = dcc_searchlist(description, user, CType, 1, (char *) 0, -1);
 	filesize = 0;
 
-	if (extra && *extra)
+	/* 
+	 * 'x' is reserved for future timestamp extensions,
+	 * start ignoring it now.
+	 */
+	if (extra && *extra && *extra != 'x')
 		Client->cksum = m_strdup(extra);
 
 	if (Client->flags & DCC_MY_OFFER)
@@ -1821,33 +1874,53 @@ void	register_dcc_offer (char *user, char *type, char *description, char *addres
  */
 void	dcc_check (fd_set *Readables)
 {
-	DCC_list	**Client;
+	DCC_list	*Client;
 	int		previous_server;
 
-	for (Client = &ClientList; *Client != NULL;Client = &(**Client).next)
+	/* Garbage collect whenever we can */
+	if (deadlist)
+		dcc_really_erase();
+
+	/* Sanity */
+	if (!Readables)
+		return;
+
+	/* Whats with all this double-pointer chicanery anyhow? */
+	for (Client = ClientList; Client; Client = Client->next)
 	{
-		if (Readables && (*Client)->read != -1 && FD_ISSET((*Client)->read, Readables))
+		/* 
+		 * There are some ways that this can happen.
+		 * Rather than track them all down and "fix" them,
+		 * I'll just conceed that this is possible and skip them.
+		 */
+		if (Client->flags & DCC_DELETE)
+		{
+			dcc_add_deadclient(Client);
+			continue;
+		}
+
+		if (Readables && Client->read != -1 && FD_ISSET(Client->read, Readables))
 		{
 			previous_server = from_server;
 			from_server = -1;
 			message_from(NULL, LOG_DCC);
 
-			switch((*Client)->flags&DCC_TYPES)
+			switch (Client->flags & DCC_TYPES)
 			{
 				case DCC_CHAT:
-					process_incoming_chat(*Client);
+					process_incoming_chat(Client);
 					break;
 				case DCC_RAW_LISTEN:
-					process_incoming_listen(*Client);
+					process_incoming_listen(Client);
 					break;
 				case DCC_RAW:
-					process_incoming_raw(*Client);
+					process_incoming_raw(Client);
 					break;
 				case DCC_FILEOFFER:
-					process_outgoing_file(*Client);
+					process_outgoing_file(Client);
 					break;
 				case DCC_FILEREAD:
-					process_incoming_file(*Client);
+					process_incoming_file(Client);
 					break;
 			}
 
@@ -1857,16 +1930,18 @@ void	dcc_check (fd_set *Readables)
 		/*
 		 * Enforce any timeouts
 		 */
-		else
+		else if (Client->flags & (DCC_MY_OFFER | DCC_THEIR_OFFER) &&
+				dcc_timeout >= 0 &&
+				(now - Client->lasttime.tv_sec > dcc_timeout))
 		{
-			if ((*Client)->flags & (DCC_MY_OFFER|DCC_THEIR_OFFER))
-			{
-				if (now - (*Client)->lasttime.tv_sec > dcc_timeout)
-				{
-					say("Auto-rejecting a dcc after [%ld] seconds", now - (*Client)->lasttime.tv_sec);
-					(*Client)->flags |= DCC_DELETE;
-				}
-			}
+			if (do_hook(DCC_LOST_LIST,"%s %s %s IDLE TIME EXCEEDED",
+				Client->user, 
+				dcc_types[Client->flags & DCC_TYPES],
+				Client->description ? 
+				 Client->description : "<any>"))
+			    say("Auto-rejecting a dcc after [%ld] seconds", 
+				now - Client->lasttime.tv_sec);
+			Client->flags |= DCC_DELETE;
 		}
 
 
@@ -1874,15 +1949,12 @@ void	dcc_check (fd_set *Readables)
 		 * This shouldnt be neccesary any more, but what the hey,
 		 * im still paranoid.
 		 */
-		if (!*Client)
+		if (!Client)
 			break;
 
-		if ((*Client)->flags & DCC_DELETE)
-			dcc_add_deadclient(*Client);
+		if (Client->flags & DCC_DELETE)
+			dcc_add_deadclient(Client);
 	}
-
-	if (deadlist)
-		dcc_really_erase();
 }
 
 
@@ -1894,15 +1966,15 @@ static	void	process_incoming_chat (DCC_list *Client)
 {
 	struct	sockaddr_in	remaddr;
 	int	sra;
-	char	tmp[DCC_BLOCK_SIZE + 1];
-	char	tmp2[DCC_BLOCK_SIZE + 1];
+	char	tmp[IO_BUFFER_SIZE + 1];
+	char	tmp2[IO_BUFFER_SIZE + 1];
 	char	*bufptr;
 	long	bytesread;
 
 	if (Client->flags & DCC_MY_OFFER)
 	{
 		sra = sizeof(struct sockaddr_in);
-		Client->write = accept(Client->read, (struct sockaddr *) &remaddr, &sra);
+		Client->write = my_accept(Client->read, (struct sockaddr *) &remaddr, &sra);
 		Client->read = new_close(Client->read);
 		if ((Client->read = Client->write) > 0)
 			new_open(Client->read);
@@ -1986,7 +2058,7 @@ static	void	process_incoming_chat (DCC_list *Client)
 
 		if (do_hook(DCC_CHAT_LIST, "%s %s", Client->user, tmp))
 		{
-			if (away_set)
+			if (get_server_away(-2))
 			{
 				strcat(tmp, "<");
 				strcat(tmp, my_ctime(time(NULL)));
@@ -2016,7 +2088,7 @@ struct	hostent		*hp;
 	const char	*Name;
 
 	sra = sizeof(struct sockaddr_in);
-	new_socket = accept(Client->read, (struct sockaddr *) &remaddr, &sra);
+	new_socket = my_accept(Client->read, (struct sockaddr *) &remaddr, &sra);
 	if (new_socket < 0)
 	{
 		yell("### DCC Error: accept() failed.  Punting.");
@@ -2059,15 +2131,15 @@ struct	hostent		*hp;
  */
 static	void		process_incoming_raw (DCC_list *Client)
 {
-	char	tmp[DCC_BLOCK_SIZE+1];
+	char	tmp[IO_BUFFER_SIZE + 1];
 	char 	*bufptr;
 	long	bytesread;
 
         bufptr = tmp;
 	switch ((int)(bytesread = dgets(bufptr, Client->read, 0)))
 	{
-	case -1:
-	{
+	    case -1:
+	    {
 		if (do_hook(DCC_RAW_LIST, "%s %s C",
 				Client->user, Client->description))
        		if (do_hook(DCC_LOST_LIST,"%s RAW %s", 
@@ -2076,10 +2148,11 @@ static	void		process_incoming_raw (DCC_list *Client)
 				Client->user, Client->description);
 
 		Client->flags |= DCC_DELETE;
-	}
-	case 0:
-	default:
-	{
+		break;
+	    }
+	    case 0:
+	    default:
+	    {
 		from_server = primary_server;	/* Colten asked for it */
 		if (bytesread > 0 && tmp[strlen(tmp) - 1] == '\n')
 			tmp[strlen(tmp) - 1] = '\0';
@@ -2088,7 +2161,7 @@ static	void		process_incoming_raw (DCC_list *Client)
 				Client->user, Client->description, tmp))
 			say("Raw data on %s from %s: %s",
 				Client->user, Client->description, tmp);
-	}
+	    }
 	}
 	return;
 }
@@ -2120,7 +2193,7 @@ static void		process_outgoing_file (DCC_list *Client)
 		 * Open up the network connection
 		 */
 		sra = sizeof(struct sockaddr_in);
-		Client->write = accept(Client->read, (SA *) &remaddr, &sra);
+		Client->write = my_accept(Client->read, (SA *) &remaddr, &sra);
 		Client->read = new_close(Client->read);
 		if ((Client->read = Client->write) < 0)
 		{
@@ -2215,6 +2288,26 @@ static void		process_outgoing_file (DCC_list *Client)
 			Client->packets_outstanding--;
 		}
 
+		if (ntohl(bytesrecvd) > Client->bytes_sent)
+		{
+yell("### WARNING!  The other peer claims to have recieved more bytes than");
+yell("### I have actually sent so far.  Please report this to ");
+yell("### jnelson@acronet.net or #epic on EFNet right away.");
+yell("### Ask the person who you sent the file to look for garbage at the");
+yell("### end of the file you just sent them.  Please inclose that ");
+yell("### information as well as the following:");
+yell("###");
+yell("###    bytesrecvd [%ld]        Client->bytes_sent [%ld]", 
+				ntohl(bytesrecvd), (long)Client->bytes_sent);
+yell("###    Client->filesize [%ld]  Client->packets_ack [%ld]",
+				Client->filesize, Client->packets_ack);
+yell("###    Client->packets_outstanding [%ld]",
+				Client->packets_outstanding);
+
+			/* And just cope with it to avoid whining */
+			Client->bytes_sent = ntohl(bytesrecvd);
+		}
+	
 		/*
 		 * If we've sent the whole file already...
 		 */
@@ -2509,37 +2602,33 @@ static void update_transfer_buffer (char *format, ...)
 }
 
 
-static void file_checksum (unsigned char cksum[], char *filename)
+static void 	file_checksum (unsigned char cksum[], char *filename)
 {
-	FILE *open_file;
-	unsigned char values[4];
-	int i;
+	int	open_file;
+const	int	chunks_per_block = 2048;
+const	int	block_size	 = chunks_per_block * 4;
+	u_char	*block;
+	u_char	*ptr;
+	int	i, j;
 
-	cksum[0] = cksum[1] = cksum[2] = cksum[3] = 0;
-
-	if (!(open_file = fopen(filename, "r")))
+	if ((open_file = open(filename, O_RDONLY, 0)) == -1)
 	{
 		yell("### Couldnt open %s: %s\n",filename, strerror(errno));
 		return;
 	}
 
-	while ((i = fread(values, sizeof(unsigned char), 4, open_file)))
-	{
-		switch (i)
-		{
-			case 1: values[1] = 0;
-			case 2: values[2] = 0;
-			case 3: values[3] = 0;
-			case 4: break;
-			default: panic("file_checksum");
-		}
+	block = alloca(block_size);
 
-		cksum[0] ^= values[0];
-		cksum[1] ^= values[1];
-		cksum[2] ^= values[2];
-		cksum[3] ^= values[3];
+	cksum[0] = cksum[1] = cksum[2] = cksum[3] = 0;
+	while ((i = read(open_file, block, block_size)))
+	{
+		while (i < block_size)
+			block[i++] = 0;
+
+		for (ptr = block, j = 0; j < block_size; j++)
+			cksum[j % 4] ^= *ptr++;
 	}
-	fclose (open_file);
+	close(open_file);
 }
 
 
@@ -2655,6 +2744,9 @@ static void dcc_getfile_resume_demanded (char *user, char *filename, char *port,
 	if (x_debug & DEBUG_DCC_XMIT)
 		yell("GOT DCC RESUME REQUEST from [%s] [%s|%s|%s]", user, filename, port, offset);
 
+	if (!strcmp(filename, "file.ext"))
+		filename = NULL;
+
 	if (!(Client = dcc_searchlist(filename, user, DCC_FILEOFFER, 0, port, 0)))
 	{
 		if (x_debug & DEBUG_DCC_XMIT)
@@ -2700,6 +2792,9 @@ static	void	dcc_getfile_resume_start (char *nick, char *filename, char *port, ch
 	if (x_debug & DEBUG_DCC_XMIT)
 		yell("GOT CONFIRMATION FOR DCC RESUME from [%s] [%s]", nick, filename);
 
+	if (!strcmp(filename, "file.ext"))
+		filename = NULL;
+
 	if (!(Client = dcc_searchlist(filename, nick, DCC_FILEREAD, 0, port, 0)))
 		return;		/* Its fake. */
 
