@@ -10,9 +10,9 @@
 #include <math.h>
 
 /* Function decls */
-static	void	TruncateAndQuote(char **, char *, int, char *);
-static	char	*alias_special_char(char **, char *, char *, char *, int *);
-static	void	do_alias_string (char, char *);
+static	void	TruncateAndQuote(char **, const char *, int, const char *);
+static	char	*alias_special_char(char **, char *, const char *, char *, int *);
+static	void	do_alias_string (char *, char *);
 
 char *alias_string = NULL;
 
@@ -99,7 +99,7 @@ static char	*lastop (char *ptr)
  * Backslash quotes any character not within [] or (),
  * which have their own quoting rules (*sigh*)
  */ 
-static	char	*next_unit (char *str, char *args, int *arg_flag, int stage)
+static	char	*next_unit (char *str, const char *args, int *arg_flag, int stage)
 #if 0
 	char	*str,			/* start of token */
 		*args;			/* the values of $0, $1, etc */
@@ -409,7 +409,6 @@ static	char	*next_unit (char *str, char *args, int *arg_flag, int stage)
 
 			/* Taken more or less from call_user_function */
 			make_local_stack(NULL);
-/*			make_local_stack("anonymous"); */
 			window_display = 0;
 			add_local_alias("FUNCTION_RETURN", empty_string);
 			window_display = display;
@@ -1244,9 +1243,12 @@ static	char	*next_unit (char *str, char *args, int *arg_flag, int stage)
  * about this at some future date. The ^ function and some fixes by
  * troy@cbme.unsw.EDU.AU (Troy Rollo) 
  */
-char	*parse_inline(char *str, char *args, int *args_flag)
+char	*parse_inline (char *str, const char *args, int *args_flag)
 {
-	return next_unit(str, args, args_flag, NU_EXPR);
+	if (x_debug & DEBUG_NEW_MATH)
+		return matheval(str, args, args_flag);
+	else
+		return next_unit(str, args, args_flag, NU_EXPR);
 }
 
 
@@ -1265,7 +1267,7 @@ char	*parse_inline(char *str, char *args, int *args_flag)
  *		portion is written back into more_text.
  *	Backslash escapes are unescaped.
  */
-char	*expand_alias	(char *string, char *args, int *args_flag, char **more_text)
+char	*expand_alias	(const char *string, const char *args, int *args_flag, char **more_text)
 {
 	char	*buffer = NULL,
 		*ptr,
@@ -1349,7 +1351,7 @@ char	*expand_alias	(char *string, char *args, int *args_flag, char **more_text)
 				ptr++;
 				break;
 			}
-			*more_text = string + (ptr - free_stuff) + 1;
+			*more_text = (char *)(string + (ptr - free_stuff) + 1);
 			*ptr = '\0'; /* To terminate the loop */
 			break;
 		    }
@@ -1416,7 +1418,7 @@ char	*expand_alias	(char *string, char *args, int *args_flag, char **more_text)
  * The args_flag is set to 1 if any of the $n, $n-, $n-m, $-m, $*, or $() 
  * is used in the alias.  Otherwise it is left unchanged.
  */
-static	char	*alias_special_char(char **buffer, char *ptr, char *args, char *quote_em, int *args_flag)
+static	char	*alias_special_char(char **buffer, char *ptr, const char *args, char *quote_em, int *args_flag)
 {
 	char	*tmp,
 		*tmp2,
@@ -1516,6 +1518,11 @@ static	char	*alias_special_char(char **buffer, char *ptr, char *args, char *quot
 
 			alias_special_char(&sub_buffer, tmp, args, 
 						quote_em, args_flag);
+
+			/* Some kind of bogus expando */
+			if (sub_buffer == NULL)
+				sub_buffer = m_strdup(empty_string);
+
 			TruncateAndQuote(buffer, sub_buffer, length, quote_em);
 			new_free(&sub_buffer);
 			new_free(&tmpsav);
@@ -1575,15 +1582,29 @@ static	char	*alias_special_char(char **buffer, char *ptr, char *args, char *quot
 		 * this requires a recursive call to io(), so you have all of
 		 * the synchronous problems when this is mixed with /wait and
 		 * /redirect and all that jazz waiting for io() to unwind.
+		 *
+		 * This has been changed to use add_wait_prompt instead of
+		 * get_line, and hopefully get_line will go away.  I dont 
+		 * expect any problems, but watch out for this anyways.
+		 *
+		 * Also note that i added ' to this, so now you can also
+		 * do $'...' that acts just like $"..." except it only
+		 * waits for the next key rather than the next return.
 		 */
 		case DOUBLE_QUOTE:
+		case '\'':
 		{
-			if ((ptr = strchr(tmp, DOUBLE_QUOTE)))
+			if ((ptr = strchr(tmp, c)))
 				*ptr++ = 0;
 
-			alias_string = (char *) 0;
-			get_line(tmp, 0, do_alias_string);
-			TruncateAndQuote(buffer, alias_string, length, quote_em);
+			alias_string = NULL;
+			add_wait_prompt(tmp, do_alias_string, NULL, 
+				(c == DOUBLE_QUOTE) ? WAIT_PROMPT_LINE
+						    : WAIT_PROMPT_KEY, 1);
+			while (!alias_string)
+				io("Input Prompt");
+
+			TruncateAndQuote(buffer, alias_string, length,quote_em);
 			new_free(&alias_string);
 			return (ptr);
 		}
@@ -1758,11 +1779,11 @@ static	char	*alias_special_char(char **buffer, char *ptr, char *args, char *quot
 			    else
 				tmp = get_variable_with_args(ptr, args, args_flag);
 
-			    if (tmp)
-			    {
-				TruncateAndQuote(buffer, tmp, length, quote_em);
-				new_free(&tmp);
-			    }
+			    if (!tmp)
+				tmp = m_strdup(empty_string);
+
+			    TruncateAndQuote(buffer, tmp, length, quote_em);
+			    new_free(&tmp);
 
 			    if (d)
 				*rest = d;
@@ -1779,7 +1800,7 @@ static	char	*alias_special_char(char **buffer, char *ptr, char *args, char *quot
  * TruncateAndQuote: This handles string width formatting and quoting for irc 
  * variables when [] or ^x is specified.  
  */
-static	void	TruncateAndQuote(char **buff, char *add, int length, char *quote_em)
+static	void	TruncateAndQuote(char **buff, const char *add, int length, const char *quote_em)
 {
 	if (!add)
 		panic("add is NULL");
@@ -1818,8 +1839,8 @@ static	void	TruncateAndQuote(char **buff, char *add, int length, char *quote_em)
 	return;
 }
 
-static void	do_alias_string (char unused, char *not_used)
+static void	do_alias_string (char *unused, char *input)
 {
-	malloc_strcpy(&alias_string, get_input());
+	malloc_strcpy(&alias_string, input);
 }
 
