@@ -58,8 +58,9 @@ static	int	dont_pause_topic = 0;
 static	int	entry_size;
 static	int	finished_help_paging = 0;
 static	FILE *	help_fp;
-static	int	help_paused_lines;
-static	char *	help_paused_topic[100];		/* 100 should be enough */
+#define HELP_PAUSED_LINES_MAX 500
+static	int	help_paused_lines = 0;
+static	char *	help_paused_topic[HELP_PAUSED_LINES_MAX]; /* Should be enough */
 static	Screen *help_screen = (Screen *) 0;
 static	int	help_show_directory = 0;
 static	char	help_topic_list[BIG_BUFFER_SIZE + 1];
@@ -82,6 +83,9 @@ static	int	show_help (Window *window, char *name)
 	int	rows = 0;
 	char	line[256];
 
+	if (!help_fp)
+		return 0;
+
 	old_to_window = to_window;
 	to_window = window ? window : current_window;
 
@@ -140,8 +144,10 @@ static	void	help_prompt (char *name, char *line)
 		finished_help_paging = 1;
 		fclose(help_fp);
 		help_fp = NULL;
+#if 0
 		set_help_screen((Screen *) 0);
 		return;
+#endif
 	}
 
 	if (show_help(help_window, name))
@@ -248,6 +254,8 @@ static 	void 	help_pause_add_line (char *format, ...)
 	va_start (args, format);
 	vsnprintf(buf, BIG_BUFFER_SIZE - 1, format, args);
 	va_end (args);
+	if (help_paused_lines >= HELP_PAUSED_LINES_MAX)
+		panic("help_pause_add_line: would overflow the buffer");
 	malloc_strcpy(&help_paused_topic[help_paused_lines], buf);
 	help_paused_lines++;
 }
@@ -307,7 +315,7 @@ static	void	help_show_paused_topic (char *name, char *line)
 			set_help_screen((Screen *) 0);
 
 		dont_pause_topic = 0;
-		help_paused_lines = 0;
+		help_paused_lines = 0;	/* Probably should reset this ;-) */
 		i = 0;
 	}
 }
@@ -334,7 +342,6 @@ static	void	help_me (char *topics, char *args)
 	char	tmp[BIG_BUFFER_SIZE+1];
 	char	buffer[BIG_BUFFER_SIZE+1];
 	char *	pattern = NULL;
-	int	cwd;
 
 	strcpy(help_topic_list, topics);
 	ptr = get_string_var(HELP_PATH_VAR);
@@ -416,14 +423,29 @@ static	void	help_me (char *topics, char *args)
 		/*
 		 * Gather up the names of the files in the help directory.
 		 */
-		cwd = open(".", O_RDONLY);
-		chdir(path);
-		pattern = alloca(strlen(path) + 2 + strlen(this_arg) + 3);
-		strcpy(pattern, this_arg);
-		strcat(pattern, "*");
-		bsd_glob(pattern, 0 /* GLOB_MARK */, NULL, &g);
-		fchdir(cwd);
-		close(cwd);
+		{
+#ifndef HAVE_FCHDIR
+			char 	opath[MAXPATHLEN + 1];
+			getcwd(opath, MAXPATHLEN);
+#else
+			int 	cwd = open(".", O_RDONLY);
+#endif
+
+			chdir(path);
+			pattern = alloca(strlen(path) + 2 + 
+					 strlen(this_arg) + 3);
+			strcpy(pattern, this_arg);
+			strcat(pattern, "*");
+			bsd_glob(pattern, 0 /* GLOB_MARK */, NULL, &g);
+
+#ifndef HAVE_FCHDIR
+			chdir(opath);
+#else
+			fchdir(cwd);
+			close(cwd);
+#endif
+		}
+
 		for (i = 0; i < g.gl_matchc; i++)
 		{
 			char	*tmp = g.gl_pathv[i];
